#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy.cleaning.country_data import CountryData


class JHUData(CleaningBase):
    """
    Data cleaning of JHU-style dataset.
    """

    def __init__(self, filename):
        super().__init__(filename)

    def cleaned(self, **kwargs):
        """
        Return the cleaned dataset.

        Note:
            Cleaning method is defined by self._cleaning() method.

        Args:
            kwargs: keword arguments will be ignored.

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        if "population" in kwargs.keys():
            raise ValueError(
                "@population was removed in JHUData.cleaned(). Please use JHUData.subset()")
        df = self._cleaned_df.copy()
        df = df.loc[:, self.COLUMNS]
        return df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.
        This method overwrite super()._cleaning() method.

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - ISO3 (str): ISO3 code
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        df = super()._cleaning()
        # Rename the columns
        df = df.rename(
            {
                "ObservationDate": self.DATE,
                "Country/Region": self.COUNTRY,
                "Province/State": self.PROVINCE,
                "Deaths": self.F
            },
            axis=1
        )
        # ISO3 code
        if self.ISO3 not in df.columns:
            df[self.ISO3] = self.UNKNOWN
        # Confirm the expected columns are in raw data
        expected_cols = [
            self.DATE, self.ISO3, self.COUNTRY, self.PROVINCE, self.C, self.F, self.R
        ]
        self.ensure_dataframe(df, name="the raw data", columns=expected_cols)
        # Datetime columns
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Country
        df[self.COUNTRY] = df[self.COUNTRY].replace(
            {
                "Mainland China": "China",
                "Hong Kong SAR": "Hong Kong",
                "Taipei and environs": "Taiwan",
                "Iran (Islamic Republic of)": "Iran",
                "Republic of Korea": "South Korea",
                "Republic of Ireland": "Ireland",
                "Macao SAR": "Macau",
                "Russian Federation": "Russia",
                "Republic of Moldova": "Moldova",
                "Taiwan*": "Taiwan",
                "Cruise Ship": "Others",
                "United Kingdom": "UK",
                "Viet Nam": "Vietnam",
                "Czechia": "Czech Republic",
                "St. Martin": "Saint Martin",
                "Cote d'Ivoire": "Ivory Coast",
                "('St. Martin',)": "Saint Martin",
                "Congo (Kinshasa)": "Congo",
                "Congo (Brazzaville)": "Congo",
                "Congo, the Democratic Republic of the": "Congo",
                "The, Bahamas": "Bahamas",
            }
        )
        # Province
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.UNKNOWN).replace(
            {
                "Cruise Ship": "Diamond Princess",
                "Diamond Princess cruise ship": "Diamond Princess"
            }
        )
        df.loc[df[self.COUNTRY] == "Diamond Princess", [
            self.COUNTRY, self.PROVINCE]] = ["Others", "Diamond Princess"]
        # Values
        df[self.CI] = df[self.C] - df[self.F] - df[self.R]
        df[self.VALUE_COLUMNS] = df[self.VALUE_COLUMNS].astype(np.int64)
        df = df.loc[:, [self.ISO3, *self.COLUMNS]].reset_index(drop=True)
        return df

    def replace(self, country_data):
        """
        Replace a part of cleaned dataset with a dataframe.

        Args:
            country_data (covsirphy.CountryData): dataset object of the country

        Returns:
            self
        """
        self.ensure_instance(country_data, CountryData, name="country_data")
        # Read new dataset
        country = country_data.country
        new = country_data.cleaned()
        new[self.ISO3] = self.country_to_iso3(country)
        # Remove the data in the country from JHU dataset
        df = self._cleaned_df.copy()
        df = df.loc[df[self.COUNTRY] != country, :]
        # Combine JHU data and the new data
        df = pd.concat([df, new], axis=0, sort=False)
        self._cleaned_df = df.copy()
        return self

    def subset(self, country, province=None,
               start_date=None, end_date=None, population=None):
        """
        Return the subset of dataset.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            population (int or None): population value

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases, if calculated

        Notes:
            If @population is not None, the number of susceptible cases will be calculated.
            Records with Recovered > 0 will be selected.
        """
        # Subset with area and start/end date
        subset_df = super().subset(
            country=country, province=province, start_date=start_date, end_date=end_date)
        # Select records where Recovered > 0
        df = subset_df.loc[subset_df[self.R] > 0, :]
        df = df.reset_index(drop=True)
        if df.empty:
            series = subset_df[self.DATE]
            start_date = start_date or series.min().strftime(self.DATE_FORMAT)
            end_date = end_date or series.max().strftime(self.DATE_FORMAT)
            s1 = "Records with Recovered > 0 are not registered."
            s2 = f"(country={country}, province={province}, period={start_date}-{end_date})"
            raise ValueError(f"{s1} {s2}")
        # Calculate Susceptible if population value was applied
        if population is None:
            return df
        population = self.ensure_natural_int(population, name="population")
        df.loc[:, self.S] = population - df.loc[:, self.C]
        return df

    def to_sr(self, country, province=None,
              start_date=None, end_date=None, population=None):
        """

        Args:
            country (str): country name
            province (str): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            population (int): population value

        Returns:
            (pandas.DataFrame)
                Index:
                    Date (pd.TimeStamp): Observation date
                Columns:
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases

        Notes:
            @population must be specified.
            Records with Recovered > 0 will be used.
        """
        population = self.ensure_natural_int(population, "population")
        subset_df = self.subset(
            country=country, province=province,
            start_date=start_date, end_date=end_date, population=population
        )
        return subset_df.set_index(self.DATE).loc[:, [self.R, self.S]]

    @classmethod
    def from_dataframe(cls, dataframe):
        """
        Create JHUData instance using a pandas dataframe.

        Args:
            dataframe (pd.DataFrame): Cleaned dataset

        Returns:
            (covsirphy.JHUData): JHU-style dataset
        """
        instance = cls(filename=None)
        instance._cleaned_df = cls.ensure_dataframe(
            dataframe, name="dataframe", columns=cls.COLUMNS)
        return instance

    def total(self):
        """
        Return a dataframe to show chronological change of number and rates.

        Returns:
            (pandas.DataFrame): group-by Date, sum of the values

                Index:
                    Date (pd.TimeStamp): Observation date
                Columns:
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Fatal per Confirmed (int)
                    - Recovered per Confirmed (int)
                    - Fatal per (Fatal or Recovered) (int)
        """
        df = super().total()
        total_series = df.sum(axis=1)
        r_cols = self.RATE_COLUMNS[:]
        df[r_cols[0]] = df[self.F] / total_series
        df[r_cols[1]] = df[self.R] / total_series
        df[r_cols[2]] = df[self.F] / (df[self.F] + df[self.R])
        return df.loc[:, [*self.VALUE_COLUMNS, *r_cols]]

    def countries(self):
        """
        Return names of countries where records with Recovered > 0 are registered.

        Returns:
            (list[str]): list of country names
        """
        df = self._cleaned_df.copy()
        df = df.loc[df[self.R] > 0, :]
        return list(df[self.COUNTRY].unique())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate
from covsirphy.util.plotting import line_plot, box_plot
from covsirphy.cleaning.term import Term
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.population import PopulationData
from covsirphy.ode.mbase import ModelBase
from covsirphy.phase.phase_unit import PhaseUnit
from covsirphy.phase.phase_series import PhaseSeries
from covsirphy.phase.phase_estimator import MPEstimator


class Scenario(Term):
    """
    Scenario analysis.

    Args:
        jhu_data (covsirphy.JHUData): object of records
        population_data (covsirphy.PopulationData): PopulationData object
        country (str): country name
        province (str or None): province name
        tau (int or None): tau value
    """

    def __init__(self, jhu_data, population_data, country, province=None, tau=None):
        # Population
        population_data = self.ensure_instance(
            population_data, PopulationData, name="population_data")
        self.population = population_data.value(country, province=province)
        # Records
        self.jhu_data = self.ensure_instance(
            jhu_data, JHUData, name="jhu_data")
        # Area name
        self.country = country
        self.province = province or self.UNKNOWN
        self.area = JHUData.area_name(country, province)
        # First/last date of the area
        df = jhu_data.subset(country=self.country, province=self.province)
        self.subset_df = df.copy()
        self._first_date = df[self.DATE].min().strftime(self.DATE_FORMAT)
        self._last_date = df[self.DATE].max().strftime(self.DATE_FORMAT)
        # tau value must be shared
        self.tau = self.ensure_tau(tau)
        # {scenario_name: PhaseSeries}
        self._init_phase_series()

    def __getitem__(self, key):
        if key in self._series_dict:
            return self._series_dict[key]
        raise KeyError(f"{key} scenario is not registered.")

    def __setitem__(self, key, value):
        self.ensure_instance(value, PhaseSeries, name="value")
        self._series_dict[key] = value

    def _init_phase_series(self):
        """
        Initialize dictionary of phase series.
        """
        self._series_dict = {
            self.MAIN: PhaseSeries(
                self._first_date, self._last_date, self.population
            )
        }
        self.record_df = self.jhu_data.subset(
            country=self.country,
            province=self.province,
            start_date=self._first_date,
            end_date=self._last_date,
            population=self.population
        )

    @property
    def first_date(self):
        """
        str: the first date of the records
        """
        return self._first_date

    @first_date.setter
    def first_date(self, date):
        self.ensure_date_order(self._first_date, date, name="date")
        self.ensure_date_order(date, self._last_date, name="date")
        self._first_date = date
        self._init_phase_series()

    @property
    def last_date(self):
        """
        str: the last date of the records
        """
        return self._last_date

    @last_date.setter
    def last_date(self, date):
        self.ensure_date_order(self._first_date, date, name="date")
        self.ensure_date_order(date, self._last_date, name="date")
        self._last_date = date
        self._init_phase_series()

    def records(self, show_figure=True, filename=None):
        """
        Return the records as a dataframe.

        Args:
            show_figure (bool): if True, show the records as a line-plot.
            filename (str): filename of the figure, or None (show figure)

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases (> 0)

        Notes:
            Records with Recovered > 0 will be selected.
        """
        df = self.jhu_data.subset(
            country=self.country, province=self.province,
            start_date=self._first_date, end_date=self._last_date)
        if not show_figure:
            return df
        line_plot(
            df.set_index(self.DATE).drop(self.C, axis=1),
            f"{self.area}: Cases over time",
            y_integer=True,
            filename=filename
        )
        return df

    @deprecate(old="Scenario.add_phase()", new="Scenario.add()")
    def add_phase(self, **kwargs):
        return self.add(**kwargs)

    def add(self, name="Main", end_date=None, days=None,
            population=None, model=None, **kwargs):
        """
        Add a new phase.
        The start date will be the next date of the last registered phase.

        Args:
            name (str): phase series name, 'Main' or user-defined name
            end_date (str): end date of the new phase
            days (int): the number of days to add
            population (int or None): population value of the start date
            model (covsirphy.ModelBase or None): ODE model
            kwargs: keyword arguments of ODE model parameters, not including tau value.

        Returns:
            covsirphy.Scenario: self

        Notes:
            - If the phases series has not been registered, new phase series will be created.
            - Either @end_date or @days must be specified.
            - If @end_date and @days are None, the end date will be the last date of the records.
            - If both of @end_date and @days were specified, @end_date will be used.
            - If @popultion is None, initial value will be used.
            - If @model is None, the model of the last phase will be used.
            - Tau will be fixed as the last phase's value.
            - kwargs: Default values are the parameter values of the last phase.
        """
        self._ensure_name(name)
        self._series_dict[name].add(
            end_date=end_date, days=days, population=population,
            model=model, tau=self.tau, **kwargs
        )
        return self

    def _ensure_name(self, name):
        """
        Ensure that the phases series is registered.
        If not registered, copy the main series.

        Args:
            name (str): phase series name
        """
        if name in self._series_dict.keys():
            return self._series_dict[name]
        # Phase series
        series = copy.deepcopy(self._series_dict[self.MAIN])
        series.clear(include_past=False)
        self._series_dict[name] = series
        return series

    def clear(self, name="Main", include_past=False):
        """
        Clear phase information.

        Args:
            name (str): phase series name
                - if 'Main', main phase series will be used
                - if not registered, new phaseseries will be created
            include_past (bool):
                - if True, include past phases.
                - future phase are always included

        Returns:
            covsirphy.Scenario: self
        """
        self._ensure_name(name)
        self._series_dict[name].clear(include_past=include_past)
        return self

    def delete(self, phases=None, name="Main"):
        """
        Delete a phase. The phase will be combined to the previous phase.

        Args:
            phase (list[str] or None): phase names, or ['last']
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self

        Notes:
            If @phases is None, the phase series will be deleted.
            When @phase is '0th', disable 0th phase. 0th phase will not be deleted.
        """
        self._ensure_name(name)
        # Clear main series or delete sub phase series
        if phases is None:
            if name == self.MAIN:
                self.clear(name=name, include_past=True)
                return self
            self._series_dict.pop(name)
            return self
        # Delete phases
        if not isinstance(phases, list):
            raise TypeError("@phases mut be a list of phase names.")
        phases = sorted(phases, key=self.str2num, reverse=True)
        for phase in phases:
            self._series_dict[name].delete(phase)
        return self

    def disable(self, phases, name="Main"):
        """
        The phases will be disabled and removed from summary.

        Args:
            phase (list[str] or None): phase names
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        if not isinstance(phases, list):
            raise TypeError("@phases must be a list of phase names.")
        for phase in phases:
            self._series_dict[name].disable(phase)
        return self

    def enable(self, phases, name="Main"):
        """
        The phases will be enabled and appear in summary.

        Args:
            phase (list[str] or None): phase names
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        if not isinstance(phases, list):
            raise TypeError("@phases must be a list of phase names.")
        for phase in phases:
            self._series_dict[name].enable(phase)
        return self

    def combine(self, phases, name="Main", population=None, **kwargs):
        """
        Combine the sequential phases as one phase.
        New phase name will be automatically determined.

        Args:
            phases (list[str]): list of sequential phases
            name (str, optional): name of phase series
            population (int): population value of the start date
            kwargs: keyword arguments to save as phase information

        Raises:
            TypeError: @phases is not a list

        Returns:
            covsirphy.Scenario: self
        """
        phases = sorted(phases, key=self.str2num, reverse=True)
        self.delete(phases=phases[:-1], name=name)
        if population is not None:
            unit_old = self._series_dict[name].unit(phases[-1])
            unit = PhaseUnit(
                unit_old.start_date, unit_old.end_date, population)
            unit.set_ode(**kwargs)
            self._series_dict[name].replace(phases[-1], unit)
        return self

    def separate(self, date, name="Main", population=None, **kwargs):
        """
        Create a new phase with the change point.
        New phase name will be automatically determined.

        Args:
            date (str): change point, i.e. start date of the new phase
            name (str, optional): name of phase series
            population (int): population value of the change point
            kwargs: keyword arguments of PhaseUnit.set_ode()

        Returns:
            covsirphy.Scenario: self
        """
        series = self._ensure_name(name)
        try:
            phase, old = [
                (self.num2str(i), unit) for (i, unit) in enumerate(series) if date in unit][0]
        except IndexError:
            raise IndexError(f"Phase on @date ({date}) is not registered.")
        new_pre = PhaseUnit(
            old.start_date, self.yesterday(date), old.population)
        new_pre.set_ode(**old.to_dict())
        new_fol = PhaseUnit(date, old.end_date, population or old.population)
        new_fol.set_ode(**kwargs)
        self._series_dict[name].replaces(phase, [new_pre, new_fol])
        return self

    def _summary(self, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown

        Returns:
            (pandas.DataFrame):
            - if @name not None, as the same as PhaseSeries().summary()
            - if @name is None, index will be phase series name and phase name

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        if name is None:
            if len(self._series_dict.keys()) > 1:
                dataframes = []
                for (_name, series) in self._series_dict.items():
                    summary_df = series.summary()
                    summary_df = summary_df.rename_axis(self.PHASE)
                    summary_df[self.SERIES] = _name
                    dataframes.append(summary_df.reset_index())
                df = pd.concat(dataframes, ignore_index=True, sort=False)
                return df.set_index([self.SERIES, self.PHASE])
            name = self.MAIN
        series = self._ensure_name(name)
        return series.summary()

    def summary(self, columns=None, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown
            columns (list[str] or None): columns to show

        Returns:
            pandas.DataFrame:
            - if @name not None, as the same as PhaseSeries().summary()
            - if @name is None, index will be phase series name and phase name

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
            If @columns is None, all columns will be shown.
        """
        df = self._summary(name=name)
        all_cols = df.columns.tolist()
        if set(self.EST_COLS).issubset(set(all_cols)):
            all_cols = [col for col in all_cols if col not in self.EST_COLS]
            all_cols += self.EST_COLS
        columns = columns or all_cols
        if not isinstance(columns, list):
            raise TypeError("@columns must be None or a list of strings.")
        if not set(columns).issubset(set(df.columns)):
            raise KeyError(
                f"Un-registered columns were selected as @columns. Please use {', '.join(df.columns)}."
            )
        df = df.loc[:, columns]
        return df.dropna(how="all", axis=1).fillna(self.UNKNOWN)

    def trend(self, set_phases=True, include_init_phase=False, name="Main",
              show_figure=True, filename=None, **kwargs):
        """
        Perform S-R trend analysis and set phases.

        Args:
            set_phases (bool): if True, set phases automatically with S-R trend analysis
            include_init_phase (bool): whether use initial phase or not
            name (str): phase series name
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
            kwargs: keyword arguments of ChangeFinder()

        Returns:
            covsirphy.Scenario: self

        Notes:
            If @set_phase is True and@include_init_phase is False, initial phase will not be included.
        """
        if "n_points" in kwargs.keys():
            raise ValueError(
                "@n_points argument is un-necessary"
                " because the number of change points will be automatically determined."
            )
        self._ensure_name(name)
        sr_df = self.jhu_data.to_sr(
            country=self.country, province=self.province, population=self.population
        )
        self._series_dict[name].trend(
            sr_df=sr_df,
            set_phases=set_phases,
            area=self.area,
            show_figure=show_figure,
            filename=filename,
            **kwargs
        )
        if include_init_phase:
            self._series_dict[name].enable("0th")
        return self

    def _ensure_past_phases(self, phases=None, name="Main"):
        """
        Ensure that the phases are past phases.

        Args:
            phases (list[str]): list of phase names, like 1st, 2nd...
            name (str): phase series name

        Returns:
            list[covsirphy.PhaseUnit]: list of names of phase units

        Notes:
            If @phases is None, return the all past phases.
        """
        series = self._ensure_name(name)
        past_units = [
            unit.set_id(phase=self.num2str(num))
            for (num, unit) in enumerate(series)
            if unit and unit <= self.last_date and unit.id_dict is None
        ]
        if not past_units:
            raise ValueError(
                "Scenario.trend(), Scenario.enable() or Scenario.add() must be done in advance.")
        if phases is None:
            return past_units
        if not isinstance(phases, list):
            raise TypeError("@phases must be None or a list of phase names.")
        selected_units = [series.unit(phase) for phase in phases]
        return list(set(selected_units) & set(past_units))

    def estimate(self, model, phases=None, name="Main", n_jobs=-1, **kwargs):
        """
        Estimate the parameters of the model using the records.

        Args:
            model (covsirphy.ModelBase): ODE model
            phases (list[str]): list of phase names, like 1st, 2nd...
            name (str): phase series name
            n_jobs (int): the number of parallel jobs or -1 (CPU count)
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()

        Notes:
            - If 'Main' was used as @name, main PhaseSeries will be used.
            - If @name phase was not registered, new PhaseSeries will be created.
            - If @phases is None, all past phase will be used.
            - In kwargs, tau value cannot be included.
        """
        model = self.ensure_subclass(model, ModelBase, "model")
        units = self._ensure_past_phases(phases=phases, name=name)
        # tau value must be specified in Scenario.__init__
        if self.TAU in kwargs:
            raise ValueError(
                "@tau must be specified when scenario = Scenario(), and cannot be specified here.")
        # Parameter estimation
        mp_estimator = MPEstimator(
            record_df=self.record_df, model=model, tau=self.tau, **kwargs
        )
        mp_estimator.add(units)
        results = mp_estimator.run(n_jobs=n_jobs, **kwargs)
        self.tau = mp_estimator.tau
        # Register the results
        self._series_dict[name].replaces(
            phase=None, new_list=results, keep_old=True)

    def phase_estimator(self, phase, name="Main"):
        """
        Return the estimator of the phase.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name

        Return:
            covsirphy.Estimator: estimator of the phase
        """
        estimator = self._series_dict[name].unit(phase).estimator
        if estimator is None:
            raise AttributeError(
                f'Scenario.estimate(model, phases=["{phase}"], name={name}) must be done in advance.'
            )
        return estimator

    def estimate_history(self, phase, name="Main", **kwargs):
        """
        Show the history of optimization.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.Estimator.history()

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        estimator = self.phase_estimator(phase=phase, name=name)
        estimator.history(**kwargs)

    def estimate_accuracy(self, phase, name="Main", **kwargs):
        """
        Show the accuracy as a figure.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.Estimator.accuracy()

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        estimator = self.phase_estimator(phase=phase, name=name)
        estimator.accuracy(**kwargs)

    def simulate(self, name="Main", y0_dict=None, show_figure=True, filename=None):
        """
        Simulate ODE models with set parameter values and show it as a figure.

        Args:
            name (str): phase series name. If 'Main', main PhaseSeries will be used
            y0_dict (dict or None):
                - key (str): variable name
                - value (float): initial value
                - dictionary of initial values or None
                - if model will be changed in the later phase, must be specified
            show_figure (bool):
                - if True, show the result as a figure.
            filename (str): filename of the figure, or None (show figure)

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - variables of the models (int): Confirmed (int) etc.
        """
        series = self._ensure_name(name)
        # Simulation
        sim_df = series.simulate(record_df=self.record_df, y0_dict=y0_dict)
        if not show_figure:
            return sim_df
        # Show figure
        df = sim_df.set_index(self.DATE)
        fig_cols_set = set(df.columns) & set(self.FIG_COLUMNS)
        fig_cols = [col for col in self.FIG_COLUMNS if col in fig_cols_set]
        change_dates = [unit.start_date for unit in series._units[1:]]
        line_plot(
            df[fig_cols],
            title=f"{self.area}: Predicted number of cases ({name} scenario)",
            filename=filename,
            y_integer=True,
            v=change_dates
        )
        return sim_df

    def get(self, param, phase="last", name="Main"):
        """
        Get the parameter value of the phase.

        Args:
            param (str): parameter name (columns in self.summary())
            phase (str): phase name or 'last'
                - if 'last', the value of the last phase will be returned
            name (str): phase series name

        Returns:
            (str or int or float)

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        df = self._ensure_name(name).summary()
        if param not in df.columns:
            raise KeyError(f"@param must be in {', '.join(df.columns)}.")
        if phase == "last":
            phase = df.index[-1]
        return df.loc[phase, param]

    def _param_history(self, targets, name):
        """
        Return the subset of summary dataframe to select the target of parameter history.

        Args:
            targets (list[str] or str): parameters to show (Rt etc.)
            name (str): phase series name

        Returns:
            (pandas.DataFrame): selected summary dataframe

        Raises:
            KeyError: targets are not in the columns of summary dataframe
        """
        series = self._series_dict[name]
        model_set = {unit.model for unit in series}
        model_set = model_set - set([None])
        parameters = self.flatten([m.PARAMETERS for m in model_set])
        day_params = self.flatten([m.DAY_PARAMETERS for m in model_set])
        selectable_cols = [self.N, *parameters, self.RT, *day_params]
        selectable_set = set(selectable_cols)
        df = series.summary().replace(self.UNKNOWN, None)
        if not selectable_set.issubset(set(df.columns)):
            raise ValueError(
                f"Scenario.estimate(model, phases=None, name={name}) must be done in advance.")
        targets = [targets] if isinstance(targets, str) else targets
        targets = targets or selectable_cols
        if not set(targets).issubset(selectable_set):
            raise KeyError(
                f"@targets must be selected from {', '.join(selectable_cols)}."
            )
        df = df.loc[:, targets].dropna(how="any", axis=0)
        return df.astype(np.float64)

    def param_history(self, targets=None, name="Main", divide_by_first=True,
                      show_figure=True, filename=None, show_box_plot=True, **kwargs):
        """
        Return subset of summary and show a figure to show the history.

        Args:
            targets (list[str] or str): parameters to show (Rt etc.)
            name (str): phase series name
            divide_by_first (bool): if True, divide the values by 1st phase's values
            show_box_plot (bool): if True, box plot. if False, line plot
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
            kwargs: keyword arguments of pd.DataFrame.plot or line_plot()

        Returns:
            pandas.DataFrame

        Notes:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        self._ensure_name(name)
        # Select target to show
        df = self._param_history(targets, name)
        # Divide by the first phase parameters
        if divide_by_first:
            df = df / df.iloc[0, :]
            title = f"{self.area}: Ratio to 1st phase parameters ({name} scenario)"
        else:
            title = f"{self.area}: History of parameter values ({name} scenario)"
        if not show_figure:
            return df
        if show_box_plot:
            h_values = [1.0] if divide_by_first or self.RT in targets else None
            box_plot(df, title, h=h_values, filename=filename)
            return df
        _df = df.reset_index(drop=True)
        _df.index = _df.index + 1
        h = 1.0 if divide_by_first else None
        line_plot(
            _df, title=title,
            xlabel="Phase", ylabel=str(), math_scale=False, h=h,
            filename=filename
        )
        return df

    def describe(self, y0_dict=None, with_rt=True):
        """
        Describe representative values.

        Args:
            y0_dict (dict or None):
                - key (str): variable name
                - value (float): initial value
                - dictionary of initial values or None
                - if model will be changed in the later phase, must be specified
            with_rt (bool): whether show the history of Rt values

        Returns:
            (pandas.DataFrame)
                Index:
                    (int): scenario name
                Columns:
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Infected({date}): Infected on the end date of the last phase
                    - Fatal({date}): Fatal on the end date of the last phase
        """
        _dict = {}
        for (name, _) in self._series_dict.items():
            # Predict the number of cases
            df = self.simulate(name=name, y0_dict=y0_dict, show_figure=False)
            df = df.set_index(self.DATE)
            cols = df.columns[:]
            last_date = df.index[-1]
            # Max value of Infected
            max_ci = df[self.CI].max()
            argmax_ci = df[self.CI].idxmax().strftime(self.DATE_FORMAT)
            # Infected on the end date of the last phase
            last_ci = df.loc[last_date, self.CI]
            # Fatal on the end date of the last phase
            last_f = df.loc[last_date, self.F] if self.F in cols else None
            # Save representative values
            _dict[name] = {
                f"max({self.CI})": max_ci,
                f"argmax({self.CI})": argmax_ci,
                f"{self.CI} on {last_date.strftime(self.DATE_FORMAT)}": last_ci,
                f"{self.F} on {last_date.strftime(self.DATE_FORMAT)}": last_f,
            }
        desc_df = pd.DataFrame.from_dict(_dict, orient="index")
        # History of reproduction number
        if with_rt and len(self._series_dict) > 1:
            rt_df = self.summary().reset_index()
            rt_df = rt_df.pivot_table(
                index=self.SERIES, columns=self.PHASE, values=self.RT)
            rt_df = rt_df.fillna(self.UNKNOWN)
            rt_df = rt_df.loc[:, rt_df.nunique() > 1]
            cols = sorted(rt_df, key=self.str2num)
            desc_df = desc_df.join(
                rt_df[cols].add_suffix(f"_{self.RT}"), how="left")
        return desc_df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import numpy as np
import optuna
import pandas as pd
from covsirphy.util.argument import find_args
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.util.stopwatch import StopWatch
from covsirphy.ode.mbase import ModelBase
from covsirphy.simulation.optimize import Optimizer
from covsirphy.simulation.simulator import ODESimulator


class Estimator(Optimizer):
    """
    Hyperparameter optimization of an ODE model.

    Args:
        record_df (pandas.DataFrame)
            Index:
                reset index
            Columns:
                - Date (pd.TimeStamp): Observation date
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - any other columns will be ignored
        model (covsirphy.ModelBase): ODE model
        population (int): total population in the place
        tau (int): tau value [min], a divisor of 1440
        kwargs: parameter values of the model and data subseting
    """
    np.seterr(divide="raise")

    def __init__(self, record_df, model, population, tau=None, **kwargs):
        # Arguments
        self.population = self.ensure_population(population)
        self.model = self.ensure_subclass(model, ModelBase, name="model")
        # Dataset
        if isinstance(record_df, JHUData):
            subset_arg_dict = find_args(
                [JHUData.subset, record_df.subset], **kwargs)
            self.record_df = record_df.subset(
                population=population, **subset_arg_dict)
        else:
            if not set(self.NLOC_COLUMNS).issubset(record_df.columns):
                record_df = model.restore(record_df)
            self.record_df = self.ensure_dataframe(
                record_df, name="record_df", columns=self.NLOC_COLUMNS
            )
        # Initial values
        df = model.tau_free(self.record_df, population, tau=None)
        self.y0_dict = {
            k: df.loc[df.index[0], k] for k in model.VARIABLES
        }
        # Fixed parameter values
        self.fixed_dict = {
            k: v for (k, v) in kwargs.items()
            if k in set(model.PARAMETERS) and v is not None
        }
        # For optimization
        optuna.logging.disable_default_handler()
        self.x = self.TS
        self.y_list = model.VARIABLES[:]
        self.weight_dict = {
            v: p for (v, p) in zip(model.VARIABLES, model.WEIGHTS) if p > 0
        }
        self.study = None
        self.total_trials = 0
        self.run_time = 0
        self.tau_candidates = self.divisors(1440)
        # Defined in parent class, but not used
        self.train_df = None
        # step_n will be defined in divide_minutes()
        self.step_n = None
        # tau value
        self.tau = self.ensure_tau(tau)
        self.taufree_df = pd.DataFrame() if tau is None else self.divide_minutes(tau)

    def run(self, timeout=60, reset_n_max=3,
            timeout_iteration=5, allowance=(0.98, 1.02), seed=0, **kwargs):
        """
        Run optimization.
        If the result satisfied the following conditions, optimization ends.
        - all values are not under than 0
        - values of monotonic increasing variables increases monotonically
        - predicted values are in the allowance when each actual value shows max value

        Args:
            timeout (int): time-out of run
            reset_n_max (int): if study was reset @reset_n_max times, will not be reset anymore
            timeout_iteration (int): time-out of one iteration
            allowance (tuple(float, float)): the allowance of the predicted value
            seed (int or None): random seed of hyperparameter optimization
            kwargs: other keyword arguments will be ignored

        Notes:
            @n_jobs was obsoleted because this is not effective for Optuna.
        """
        # Create a study of optuna
        if self.study is None:
            self._init_study(seed=seed)
        reset_n = 0
        iteration_n = math.ceil(timeout / timeout_iteration)
        increasing_cols = [f"{v}{self.P}" for v in self.model.VARS_INCLEASE]
        stopwatch = StopWatch()
        for _ in range(iteration_n):
            # Perform optimization
            self.study.optimize(
                self.objective, n_jobs=1, timeout=timeout_iteration)
            # Create a table to compare observed/estimated values
            tau = self.tau or super().param()[self.TAU]
            train_df = self.divide_minutes(tau)
            comp_df = self.compare(train_df, self.predict())
            # Check monotonic variables
            mono_ok_list = [
                comp_df[col].is_monotonic_increasing for col in increasing_cols
            ]
            if not all(mono_ok_list):
                if reset_n == reset_n_max - 1:
                    break
                # Initialize the study
                self._init_study()
                reset_n += 1
                continue
            # Need additional trials when the values are not in allowance
            if self._is_in_allowance(comp_df, allowance):
                break
        # Calculate run-time and the number of trials
        self.run_time = stopwatch.stop()
        self.run_time_show = stopwatch.show()
        self.total_trials = len(self.study.trials)

    def _is_in_allowance(self, comp_df, allowance):
        """
        Return whether all max values of predicted values are in allowance or not.

        Args:
            comp_df (pandas.DataFrame): [description]
            allowance (tuple(float, float)): the allowance of the predicted value

        Returns:
            (bool): True when all max values of predicted values are in allowance
        """
        a_max_values = [comp_df[f"{v}{self.A}"].max() for v in self.y_list]
        p_max_values = [comp_df[f"{v}{self.P}"].max() for v in self.y_list]
        allowance0, allowance1 = allowance
        ok_list = [
            (a * allowance0 <= p) and (p <= a * allowance1)
            for (a, p) in zip(a_max_values, p_max_values)
        ]
        return all(ok_list)

    def objective(self, trial):
        """
        Objective function of Optuna study.
        This defines the parameter values using Optuna.

        Args:
            trial (optuna.trial): a trial of the study

        Returns:
            (float): score of the error function to minimize
        """
        # Convert T to t using tau
        taufree_df = self.taufree_df.copy()
        if taufree_df.empty:
            tau = trial.suggest_categorical(self.TAU, self.tau_candidates)
            taufree_df = self.divide_minutes(tau)
        # Set parameters of the models
        model_param_dict = self.model.param_range(
            taufree_df, self.population)
        p_dict = {
            k: trial.suggest_uniform(k, *v)
            for (k, v) in model_param_dict.items()
            if k not in self.fixed_dict.keys()
        }
        p_dict.update(self.fixed_dict)
        return self.error_f(p_dict, taufree_df)

    def divide_minutes(self, tau):
        """
        Divide T by tau in the training dataset and calculate the number of steps.

        Args:
            tau (int): tau value [min]

        Returns:
            (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - t (int): Elapsed time divided by tau value [-]
                    - columns with dimensional variables
        """
        df = self.model.tau_free(self.record_df, self.population, tau=tau)
        self.step_n = int(df[self.TS].max())
        return df

    def error_f(self, param_dict, taufree_df):
        """
        Definition of error score to minimize in the study.

        Args:
            param_dict (dict): estimated parameter values
                - key (str): parameter name
                - value (int or float): parameter value
            taufree_df (pandas.DataFrame): training dataset

                Index:
                    reset index
                Columns:
                    - t: time steps [-]
                    - columns with dimensional variables

        Returns:
            (float): score of the error function to minimize
        """
        sim_df = self.simulate(self.step_n, param_dict)
        comp_df = self.compare(taufree_df, sim_df)
        # Calculate error score
        try:
            return sum(
                self._score(variable, comp_df)
                for variable in self.weight_dict.keys()
            )
        except (ZeroDivisionError, TypeError):
            return np.inf

    def _score(self, v, comp_df):
        """
        Calculate score of the variable.

        Args:
            v (str): variable na,e
            com_df (pandas.DataFrame):
                Index:
                    (str): time step
                Columns:
                    - columns with "_actual"
                    - columns with "_predicted"
                    - columns are defined by self.y_list

        Returns:
            float: score
        """
        weight = self.weight_dict[v]
        actual = comp_df.loc[:, f"{v}{self.A}"]
        diff = (actual - comp_df.loc[:, f"{v}{self.P}"]).abs() / (actual + 1)
        return weight * diff.mean()

    def simulate(self, step_n, param_dict):
        """
        Simulate the values with the parameters.

        Args:
            step_n (int): number of iteration
            param_dict (dict): estimated parameter values
                - key (str): parameter name
                - value (int or float): parameter value

        Returns:
            (pandas.DataFrame):
                Index:
                    reset index
                Columns:
                    - t (int): Elapsed time divided by tau value [-]
                    - columns with dimensionalized variables
        """
        simulator = ODESimulator()
        simulator.add(
            model=self.model,
            step_n=step_n,
            population=self.population,
            param_dict=param_dict,
            y0_dict=self.y0_dict
        )
        return simulator.taufree()

    def to_dict(self):
        """
        Summarize the results of optimization.

        Args:
            name (str or None): index of the dataframe

        Returns:
            (pandas.DataFrame):
                Index:
                    name or reset index (when name is None)
                Columns:
                    - (parameters of the model)
                    - tau
                    - Rt: basic or phase-dependent reproduction number
                    - (dimensional parameters [day])
                    - RMSLE: Root Mean Squared Log Error
                    - Trials: the number of trials
                    - Runtime: run time of estimation
        """
        est_dict = super().param()
        if self.TAU not in est_dict:
            est_dict[self.TAU] = self.tau
        model_instance = self.model(
            population=self.population,
            **{k: v for (k, v) in est_dict.items() if k != self.TAU}
        )
        minutes, seconds = divmod(int(self.run_time), 60)
        return {
            **est_dict,
            self.RT: model_instance.calc_r0(),
            **model_instance.calc_days_dict(est_dict[self.TAU]),
            self.RMSLE: self._rmsle(est_dict[self.TAU]),
            self.TRIALS: self.total_trials,
            self.RUNTIME: f"{minutes} min {seconds:>2} sec"
        }

    def summary(self, name=None):
        """
        Summarize the results of optimization.

        Args:
            name (str or None): index of the dataframe

        Returns:
            (pandas.DataFrame):
                Index:
                    name or reset index (when name is None)
                Columns:
                    - (parameters of the model)
                    - tau
                    - Rt: basic or phase-dependent reproduction number
                    - (dimensional parameters [day])
                    - RMSLE: Root Mean Squared Log Error
                    - Trials: the number of trials
                    - Runtime: run time of estimation
        """
        summary_dict = {name or 0: self.to_dict()}
        df = pd.DataFrame.from_dict(summary_dict, orient="index")
        return df.fillna(self.UNKNOWN)

    def _rmsle(self, tau):
        """
        Return RMSLE score.

        Args:
            tau (int): tau value [min]
        """
        return super().rmsle(
            train_df=self.divide_minutes(tau),
            dim=1
        )

    def accuracy(self, show_figure=True, filename=None):
        """
        Show the accuracy as a figure.

        Args:
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
        """
        est_dict = super().param()
        if self.TAU not in est_dict:
            est_dict[self.TAU] = self.tau
        train_df = self.divide_minutes(est_dict[self.TAU])
        use_variables = [
            v for (i, (p, v))
            in enumerate(zip(self.model.WEIGHTS, self.model.VARIABLES))
            if p != 0 and i != 0
        ]
        return super().accuracy(
            train_df=train_df,
            variables=use_variables,
            show_figure=show_figure,
            filename=filename
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pandas as pd
import pytest
import warnings
from covsirphy import DataLoader
from covsirphy import Term, JHUData, CountryData, PopulationData, OxCGRTData
from covsirphy import Population


class TestJHUData(object):
    def test_jhu(self, data_loader):
        jhu_data = data_loader.jhu()
        assert data_loader.covid19dh_citation
        assert isinstance(jhu_data, JHUData)
        assert isinstance(jhu_data.citation, str)
        df = jhu_data.cleaned()
        assert isinstance(df, pd.DataFrame)
        assert set(df.columns) == set(Term.COLUMNS)

    def test_jhu_local_file(self, data_loader):
        local_path = Path("input") / "covid19dh.csv"
        data_loader.jhu(local_file=local_path)
        local_file = str(local_path)
        jhu_data = data_loader.jhu(local_file=local_file)
        assert jhu_data.citation == str()
        jhu_data.citation = "citation"
        assert jhu_data.citation == "citation"

    def test_jhu_local_file_unexpected(self, data_loader):
        local_path = Path("input") / "covid_jpn_total.csv"
        with pytest.raises(Exception):
            data_loader.jhu(local_file=local_path)

    def test_replace(self, jhu_data, japan_data):
        jhu_data.replace(japan_data)
        assert isinstance(jhu_data, JHUData)
        replaced_df = jhu_data.subset("Japan")
        japan_df = japan_data.cleaned()
        assert set(replaced_df.columns) == set(Term.NLOC_COLUMNS)
        assert len(replaced_df) == len(japan_df)

    def test_cleaned(self, jhu_data):
        with pytest.raises(ValueError):
            jhu_data.cleaned(population=10000)

    def test_iso3_to_country(self, jhu_data):
        assert jhu_data.iso3_to_country("JPN") == "Japan"
        assert jhu_data.iso3_to_country("Japan") == "Japan"
        assert jhu_data.iso3_to_country("AAA") == "AAA"

    def test_country_to_iso3(self, jhu_data):
        assert jhu_data.country_to_iso3("Japan") == "JPN"
        with pytest.raises(KeyError):
            jhu_data.country_to_iso3("Country")

    def test_area_name(self, jhu_data):
        name_country = jhu_data.area_name("Japan")
        assert name_country == "Japan"
        name_with_province = jhu_data.area_name("Japan", province="Tokyo")
        assert name_with_province == "Japan/Tokyo"

    def test_subset(self, jhu_data):
        # With country name
        df = jhu_data.subset("Japan")
        assert isinstance(df, pd.DataFrame)
        assert set(df.columns) == set(Term.NLOC_COLUMNS)
        # Province
        if "Tokyo" in jhu_data.cleaned().columns:
            df = jhu_data.subset("Japan", province="Tokyo")
            assert set(df.columns) == set(Term.NLOC_COLUMNS)
        # Date
        df = jhu_data.subset(
            "Japan", start_date="01Feb2020", end_date="01Mar2020"
        )
        assert set(df.columns) == set(Term.NLOC_COLUMNS)
        # Calculate Susceptible with population value
        df = jhu_data.subset("Japan", population=125000000)
        assert set(df.columns) == set([*Term.NLOC_COLUMNS, Term.S])

    def test_countries(self, jhu_data):
        countries = jhu_data.countries()
        assert isinstance(countries, list)
        for country in countries:
            assert len(jhu_data.subset(country)) > 0

    def test_total(self, jhu_data):
        df = jhu_data.total()
        assert set(df.columns) == set(
            [*Term.VALUE_COLUMNS, *Term.RATE_COLUMNS])

    def test_from_dataframe(self, jhu_data):
        df = jhu_data.cleaned()
        jhu_data2 = JHUData.from_dataframe(df)
        assert set(df.columns) == set(jhu_data2.cleaned().columns)


class TestJapanData(object):
    def test_japan_cases(self, data_loader):
        japan_data = data_loader.japan()
        assert isinstance(japan_data, CountryData)
        assert isinstance(japan_data.citation, str)
        df = japan_data.cleaned()
        assert isinstance(df, pd.DataFrame)
        assert set(df.columns) == set(Term.COLUMNS)

    def test_japan_cases_local_file(self, data_loader):
        local_path = Path("input") / "covid_jpn_total.csv"
        data_loader.japan(local_file=local_path)
        local_file = str(local_path)
        japan_data = data_loader.japan(local_file=local_file)
        assert japan_data.citation == str()

    def test_japan_cases_local_file_unexpected(self):
        data_loader = DataLoader("input")
        local_path = Path("input") / "covid19dh.csv"
        with pytest.raises(Exception):
            warnings.filterwarnings("ignore", category=pd.errors.DtypeWarning)
            data_loader.japan(local_file=local_path)

    def test_total(self, japan_data):
        df = japan_data.total()
        assert set(df.columns) == set(
            [*Term.VALUE_COLUMNS, *Term.RATE_COLUMNS])

    def test_countries(self, japan_data):
        countries = japan_data.countries()
        assert countries == ["Japan"]


class TestPopulationData(object):
    def test_population(self, data_loader):
        warnings.filterwarnings("error", category=FutureWarning)
        population_data = data_loader.population()
        assert isinstance(population_data, PopulationData)
        assert isinstance(population_data.citation, str)
        df = population_data.cleaned()
        assert isinstance(df, pd.DataFrame)
        assert set(df.columns) == set(PopulationData.POPULATION_COLS)
        assert isinstance(population_data.to_dict(), dict)
        assert isinstance(population_data.to_dict(country_level=False), dict)
        population_data.update(10_000, "Example")
        assert population_data.value("Example") == 10_000
        population_data.update(15_000, "Example")
        assert population_data.value("Example") == 15_000
        population_data.update(18_000, "Example", date="01Feb2020")
        assert population_data.value("Example", date="01Feb2020") == 18_000
        assert population_data.value("Example") == 15_000
        with pytest.raises(KeyError):
            population_data.value("Example", date="01Jan2020")

    def test_population_local_file(self, data_loader):
        local_path = Path("input") / "covid19dh.csv"
        data_loader.population(local_file=local_path)
        local_file = str(local_path)
        population_data = data_loader.population(local_file=local_file)
        assert population_data.citation == str()

    def test_population_local_file_unexpected(self, data_loader):
        local_path = Path("input") / "covid_jpn_total.csv"
        with pytest.raises(Exception):
            data_loader.population(local_file=local_path)

    def test_population_value(self, population_data):
        warnings.filterwarnings("error", category=FutureWarning)
        df = population_data.cleaned()
        if "JPN" in df["ISO3"].unique():
            assert isinstance(population_data.value("JPN"), int)
        else:
            with pytest.raises(KeyError):
                population_data.value("JPN")
        assert isinstance(population_data.value("Japan"), int)
        with pytest.raises(KeyError):
            population_data.value("Example")

    def test_total(self, population_data):
        value = population_data.total()
        assert isinstance(value, int)

    def test_countries(self, population_data):
        countries = population_data.countries()
        assert isinstance(countries, list)
        for country in countries:
            population_data.value(country)

    def test_population_dep(self):
        warnings.simplefilter("error")
        with pytest.raises(DeprecationWarning):
            assert Population(filename=None)
        warnings.simplefilter("ignore")
        assert Population(filename=None)


class TestOxCGRTData(object):
    def test_oxcgrt(self, data_loader):
        oxcgrt_data = data_loader.oxcgrt()
        assert isinstance(oxcgrt_data, OxCGRTData)
        assert isinstance(oxcgrt_data.citation, str)
        df = oxcgrt_data.cleaned()
        assert isinstance(df, pd.DataFrame)
        assert set(df.columns) == set(OxCGRTData.OXCGRT_COLS)
        subset_cols_set = set(OxCGRTData.OXCGRT_COLS_WITHOUT_COUNTRY)
        subset_df = oxcgrt_data.subset("Japan")
        assert set(subset_df.columns) == subset_cols_set
        subset_df_iso = oxcgrt_data.subset("JPN")
        assert set(subset_df_iso.columns) == subset_cols_set
        with pytest.raises(KeyError):
            assert oxcgrt_data.subset("Moon").empty

    def test_oxcgrt_local_file(self, data_loader):
        local_path = Path("input") / "covid19dh.csv"
        data_loader.oxcgrt(local_file=local_path)
        local_file = str(local_path)
        oxcgrt_data = data_loader.oxcgrt(local_file=local_file)
        assert oxcgrt_data.citation == str()

    def test_oxcgrt_local_file_unexpected(self, data_loader):
        local_path = Path("input") / "covid_jpn_total.csv"
        with pytest.raises(Exception):
            data_loader.oxcgrt(local_file=local_path)

    def test_total(self, oxcgrt_data):
        with pytest.raises(NotImplementedError):
            oxcgrt_data.total()

    def test_countries(self, oxcgrt_data):
        countries = oxcgrt_data.countries()
        assert isinstance(countries, list)
        for country in countries:
            assert len(oxcgrt_data.subset(country)) > 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
try:
    import better_exceptions
    better_exceptions_installed = True
except ImportError:
    better_exceptions_installed = False
from covsirphy.__version__ import __version__
from covsirphy.cleaning import Term, CleaningBase, DataLoader
from covsirphy.cleaning import JHUData, CountryData, PopulationData, OxCGRTData
from covsirphy.ode import ModelBase, SIR, SIRD, SIRF, SIRFV, SEWIRF
from covsirphy.phase import Trend, ChangeFinder, PhaseSeries, PhaseUnit, MPEstimator
from covsirphy.simulation import Optimizer, ODESimulator, Estimator
from covsirphy.util import line_plot, jpn_map, StopWatch, error, find_args
from covsirphy.analysis import ExampleData, Scenario
from covsirphy.worldwide import PolicyMeasures
# Deprecated
from covsirphy.cleaning import Population, Word


def get_version():
    """
    Return the version number, like CovsirPhy v0.0.0
    """
    return f"CovsirPhy v{__version__}"


__all__ = [
    "ODESimulator", "ChangeFinder",
    "PhaseSeries", "Scenario", "ExampleData", "PhaseUnit", "MPEstimator",
    "Term", "CleaningBase", "DataLoader",
    "JHUData", "CountryData", "PopulationData", "OxCGRTData",
    "ModelBase", "SIR", "SIRD", "SIRF", "SIRFV", "SEWIRF",
    "Estimator", "Trend", "Optimizer",
    "line_plot", "jpn_map", "StopWatch", "error", "find_args",
    "PolicyMeasures",
    # Deprecated
    "Population", "Word",
]

# Show exceptions in better format if used from command line
if not hasattr(sys, "ps1") or not better_exceptions_installed:
    better_exceptions.MAX_LENGTH = None
    better_exceptions.hook()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import warnings
from covsirphy import ChangeFinder, Trend


class TestChangeFinder(object):
    @pytest.mark.parametrize(
        "country",
        ["Italy", "Japan", "United States", "India", "New Zealand"]
    )
    def test_find(self, jhu_data, population_data, country):
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(country=country, population=population)
        change_finder = ChangeFinder(sr_df)
        change_finder.run()
        warnings.filterwarnings("ignore", category=UserWarning)
        change_finder.show(area=jhu_data.area_name(country="Italy"))
        assert isinstance(change_finder.date_range(), tuple)

    def test_find_with_small_min_size(self, jhu_data, population_data):
        population = population_data.value("Italy")
        sr_df = jhu_data.to_sr(country="Italy", population=population)
        with pytest.raises(ValueError):
            change_finder = ChangeFinder(sr_df, min_size=2)
            change_finder.run()

    def test_find_with_few_records(self, jhu_data, population_data):
        population = population_data.value("Italy")
        sr_df = jhu_data.to_sr(
            country="Italy", population=population, end_date="24Feb2020")
        with pytest.raises(ValueError):
            min_size = 7
            change_finder = ChangeFinder(sr_df, min_size=min_size)
            change_finder.run()


class TestTrend(object):
    def test_one_phase(self, jhu_data, population_data):
        population = population_data.value("Italy")
        sr_df = jhu_data.to_sr(
            country="Italy", population=population,
            start_date="25Mar2020", end_date="02Apr2020"
        )
        trend = Trend(sr_df)
        trend.run()
        assert isinstance(trend.rmsle(), float)
        warnings.filterwarnings("ignore", category=UserWarning)
        trend.show(area=jhu_data.area_name(country="Italy"))

    def test_one_phase_with_few_records(self, jhu_data, population_data):
        population = population_data.value("Italy")
        sr_df = jhu_data.to_sr(
            country="Italy", population=population,
            start_date="25Mar2020", end_date="26Mar2020"
        )
        with pytest.raises(ValueError):
            trend = Trend(sr_df)
            trend.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dask import dataframe as dd
import pandas as pd
from covsirphy.cleaning.term import Term


class CleaningBase(Term):
    """
    Basic class for data cleaning.

    Args:
        filename (str): CSV filename of the dataset
    """

    def __init__(self, filename):
        if filename is None:
            self._raw = pd.DataFrame()
            self._cleaned_df = pd.DataFrame()
        else:
            self._raw = dd.read_csv(
                filename, dtype={"Province/State": "object"}
            ).compute()
            self._cleaned_df = self._cleaning()
        self._citation = str()

    @property
    def raw(self):
        """
        Return the raw data.

        Returns:
            (pandas.DataFrame): raw data
        """
        return self._raw

    def cleaned(self):
        """
        Return the cleaned dataset.

        Notes:
            Cleaning method is defined by self._cleaning() method.

        Returns:
            (pandas.DataFrame): cleaned data
        """
        return self._cleaned_df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Notes:
            Cleaning method is defined by self._cleaning() method.

        Returns:
            (pandas.DataFrame): cleaned data
        """
        return self._raw.copy()

    @property
    def citation(self):
        """
        str: citation/description of the dataset
        """
        return self._citation

    @citation.setter
    def citation(self, description):
        self._citation = str(description)

    def iso3_to_country(self, iso3_code):
        """
        Convert ISO3 code to country name if records are available.

        Args:
            iso3_code (str): ISO3 code or country name

        Returns:
            (str): country name

        Notes:
            If ISO3 codes are not registered, return the string as-si @iso3_code.
        """
        df = self._cleaned_df.copy()
        if self.ISO3 not in df.columns or iso3_code not in df[self.ISO3].unique():
            return iso3_code
        country_dict = df.set_index(self.ISO3)[self.COUNTRY].to_dict()
        return country_dict[iso3_code]

    def country_to_iso3(self, country):
        """
        Convert country name to ISO3 code if records are available.

        Args:
            country (str): country name

        Raises:
            KeyError: ISO3 code of the country is not registered

        Returns:
            (str): ISO3 code
        """
        if self.ISO3 not in self._cleaned_df.columns:
            return self.UNKNOWN
        df = self._cleaned_df.copy()
        iso3_dict = df.set_index(self.COUNTRY)[self.ISO3].to_dict()
        if country not in iso3_dict.keys():
            raise KeyError(f"@country {country} has not been registered.")
        return iso3_dict[country]

    @classmethod
    def area_name(cls, country, province=None):
        """
        Return area name of the country/province.

        Args:
            country (str): country name or ISO3 code
            province (str): province name

        Returns:
            (str): area name

        Notes:
            If province is None or '-', return country name.
            If not, return the area name, like 'Japan/Tokyo'
        """
        if province in [None, cls.UNKNOWN]:
            return country
        return f"{country}{cls.SEP}{province}"

    def _subset_by_country(self, country):
        """
        Return the subset of the country.

        Args:
            country (str): country name or ISO3 code

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    without ISO3 column and Country column
        """
        country = self.iso3_to_country(country)
        df = self._cleaned_df.copy()
        if self.COUNTRY not in df.columns:
            return df
        df = df.loc[df[self.COUNTRY] == country, :]
        df = df.drop(self.COUNTRY, axis=1)
        df = df.drop(self.ISO3, axis=1, errors="ignore")
        return df.reset_index(drop=True)

    def _subset_by_province(self, record_df, province):
        """
        Return subset of the province.

        Args:
            record_df (pandas.DataFrame): dataframe of the records
                Index:
                    reset index
                Columns:
                    without ISO3 column and Country column
            province (str or None): province name

        Raises:
            KeyError: records of the area (country/province) are not registered

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    without ISO3, Country and Province column
        """
        province = province or self.UNKNOWN
        df = record_df.copy()
        p_series = df[self.PROVINCE]
        # Province was selected
        if province != self.UNKNOWN:
            df = df.loc[p_series == province, :]
            if df.empty:
                raise KeyError(
                    f"@province {province} has not been registered.")
            if self.DATE in df.columns:
                df = df.groupby(self.DATE).last().reset_index()
            return df.drop(self.PROVINCE, axis=1)
        # Calculate total values at country level if not registered
        total_df = df.loc[p_series != self.UNKNOWN, :]
        if self.DATE in df.columns:
            total_df = total_df.groupby(self.DATE).sum().reset_index()
            if not total_df.empty:
                total_df.loc[:, self.PROVINCE] = self.UNKNOWN
        else:
            sum_dict = total_df.sum(axis=0, numeric_only=True).to_dict()
            sum_dict[self.PROVINCE] = self.UNKNOWN
            total_df = total_df.append(pd.Series(sum_dict), ignore_index=True)
        df = pd.concat([df, total_df], axis=0, ignore_index=True, sort=False)
        if self.DATE in df.columns:
            df = df.groupby([self.PROVINCE, self.DATE]).max().reset_index()
        else:
            df = df.groupby(self.PROVINCE).max().reset_index()
        # Return country-level records
        df = df.loc[df[self.PROVINCE] == self.UNKNOWN, :]
        df = df.reset_index(drop=True)
        return df.drop(self.PROVINCE, axis=1)

    def _subset_by_area(self, country, province=None):
        """
        Return subset of the country/province.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name

        Raises:
            ValueError: @province is not None, but the dataset does not have Province column
            KeyError: records of the area (country/province) are not registered

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    without ISO3, Country, Province column
        """
        # Country level
        df = self._subset_by_country(country=country)
        if df.empty:
            raise KeyError(
                f"Please register population value for country={country}"
            )
        # Province level
        if self.PROVINCE not in df.columns:
            raise ValueError(
                "@province was specified, but the dataset does not have Province column.")
        df = self._subset_by_province(df, province=province)
        if df.empty:
            raise KeyError(
                f"Please register population value for (country={country}, province={province})"
            )
        return df

    def subset(self, country, province=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Raises:
            ValueError:
                - @country was None
                - @province is not None, but the dataset does not have Province column
            KeyError: selected records are not registered

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    without ISO3, Country, Province column
        """
        if country is None:
            ValueError("@country must be specified.")
        df = self._subset_by_area(country=country, province=province)
        # Subset with Start/end date
        if start_date is None and end_date is None:
            return df.reset_index(drop=True)
        if self.DATE not in df.columns:
            raise KeyError(
                "@start_date or @end_date was specified, but the dataset does not have Date column.")
        series = df[self.DATE].copy()
        start_obj = self.date_obj(date_str=start_date, default=series.min())
        end_obj = self.date_obj(date_str=end_date, default=series.max())
        df = df.loc[(start_obj <= series) & (series <= end_obj), :]
        df = df.reset_index(drop=True)
        if df.empty:
            s1 = f"Records from {start_date} to {end_date} were not registered."
            raise KeyError(
                f"{s1} (country={country}, province={province})")
        return df

    def countries(self):
        """
        Return names of countries where records are registered.

        Raises:
            KeyError: Country names are not registered in this dataset

        Returns:
            (list[str]): list of country names
        """
        df = self._cleaned_df.copy()
        if self.COUNTRY not in df.columns:
            raise KeyError("Country names are not registered in this dataset.")
        return list(df[self.COUNTRY].unique())

    def total(self):
        """
        Calculate total values of the cleaned dataset.

        Returns:
            (pandas.DataFrame or pandas.Series or float):
                Index: 'Date' (pandas.TimeStamp) or not exist (pandas.Series)
                Columns: column names of the cleaned dataset (dtype=int or float)
                Values: total values
        """
        df = self._cleaned_df.copy()
        cols = list(set([self.COUNTRY, self.DATE]) & set(df.columns))
        if not cols:
            return df.sum(axis=0)
        # Calculate total values at country level if not registered
        c_level_df = df.groupby(cols).sum().reset_index()
        if self.PROVINCE in df.columns:
            c_level_df[self.PROVINCE] = self.UNKNOWN
            cols = [*cols, self.PROVINCE]
        df = pd.concat([df, c_level_df], axis=0, ignore_index=True, sort=False)
        df = df.drop_duplicates(subset=cols)
        if self.PROVINCE in df.columns:
            df = df.loc[df[self.PROVINCE] == self.UNKNOWN, :]
        # Calculate total value of each column
        if self.DATE in df.columns:
            return df.groupby(self.DATE).sum()
        return df.sum()
